import numpy as np
from abc import ABC, abstractmethod


class Layer(ABC):
    '''
    Abstract class representing a neural network layer
    '''

    @abstractmethod
    def forward(self, X, train=True):
        '''
        Calculates a forward pass through the layer.

        Args:
            X (numpy.ndarray): Input  (batch_size, input_size)
            train (bool): If true caches values required for backward function

        Returns:
            Out (numpy.ndarray): Output (batch_size, output_size)
        '''

    @abstractmethod
    def backward(self, dY):
        '''
        Calculates a backward pass through the layer.

        Args:
            dY (numpy.ndarray): Output grad (batch_size, output_size)

        Returns:
            dX (numpy.ndarray): Gradient of the input (batch_size, output_size)
            var_grad_list (List[Tuple[np.array]]): variables and grads
        '''

    @staticmethod
    def _initializer_(W, init_method):
        """

        Initializes the parameter passes as argument using
        Xavier or He initialization

        Args:
            W (numpy.ndarray): Parameter to be initialized
            init_method (str): Method to initialize the parameter

        """
        init_method = init_method.lower()
        if init_method not in ['xavier', 'he']:
            raise NotImplementedError('method not currently supported')

        if len(W.shape) == 2:  # linear layer
            input_dim, output_dim = W.shape
            if init_method == 'xavier':
                return np.sqrt(2.0/(input_dim+output_dim))
            if init_method == 'he':
                return np.sqrt(2.0/(input_dim))
        if len(W.shape) == 4:  # convolutional layer
            n_filter, d_filter, h_filter, w_filter = W.shape
            return np.sqrt(2.0/(h_filter*w_filter*d_filter))

        raise NotImplementedError('This W size is not defined')

    def get_params(self):
        '''
        Returns the list of numpy array of weights
        '''
        return self.params

    def set_params(self, params):
        '''
        Sets the params of a layer with a given list of numpy arrays

        Ags:
            :params (list of numpy.ndarray): new weights

        '''
        old_params = self.get_params()
        assert len(old_params) == len(params), "length mismatch"
        assert all(params[i].shape == old_params[i].shape for i in range(
            len(old_params))), "shape missmatch"
        self.params = params.copy()

    def set_config(self, config):
        self.__init__(*config)

    def save_layer(self):
        return (('conf:', self.get_config()), ('params:', self.get_params()))

    def get_config(self):
        return {}

    def __repr__(self):
        if hasattr(self, 'l_name'):
            return f'{self.l_name} layer'
        return f'Layer'

    def __call__(self, X, train=True):
        return self.forward(X, train)



from simflow.layers.layer_class import Layer

import numpy as np
from unittest import TestCase


class HelperClass(Layer):
    def __init__(self, config, params):
        self.config = config
        self.params = params

    def forward(self, x, train):
        pass

    def backward(self):
        """dummy
        """


class TestLayerClass(TestCase):
    def setUp(self):
        self.param = np.random.random((5, 6))
        self.helper = HelperClass("config", self.param)

    def test_misc(self):
        # not really tests, for test coverage only
        assert all(self.helper.get_params().ravel() == self.param.ravel())
        self.assertEqual(self.helper.get_config(), {})
        new_params = np.random.random((5, 6))
        self.helper.set_config(["hello", new_params])
        assert all(self.helper.get_params().ravel() == new_params.ravel())
        self.assertEqual(self.helper.config, "hello")

        self.assertEqual(self.helper.__repr__(), "Layer")

        self.helper.l_name = "test"

        self.assertEqual(self.helper.__repr__(), 'test layer')

        self.helper.set_params(np.random.random((5, 6)))
        self.helper(1, 1)
        self.helper.save_layer()

    def test_construction(self):
        with self.assertRaises(TypeError):
            Layer()

    def test_init(self):
        # test initializers
        w1 = np.random.random((5, 6))
        exp_res_xavier = np.sqrt(2.0/sum(w1.shape))
        exp_res_he = np.sqrt(2.0/w1.shape[0])
        res_xavier = Layer._initializer_(w1, 'Xavier')
        res_he = Layer._initializer_(w1, 'He')
        self.assertEqual(res_xavier, exp_res_xavier)
        self.assertEqual(res_he, exp_res_he)

        with self.assertRaises(NotImplementedError):
            Layer._initializer_(w1, "randomstring")

        w2 = np.random.random((5, 6, 7, 3))
        exp_res = np.sqrt(2.0/np.product(w2.shape[1:]))
        res_ = Layer._initializer_(w2, 'Xavier')

        self.assertEqual(exp_res, res_)

        with self.assertRaises(NotImplementedError):
            Layer._initializer_(np.random.random((3, 4, 5)), 'Xavier')



import numpy as np


class Iterator:
    '''
    Class representing Iterator

    Args:
        :batch_size (int): batch size to used
        :shuffle (bool): shuffles each time if set to True
    '''

    def __init__(self, batch_size=128, *, shuffle=True, method='direct'):
        '''
        Initializes the Iterator
        Args:
            batch_size (int) : batch size to used
            shuffle (bool)   : shuffles each time if set to True
        '''
        assert isinstance(batch_size, int), f'batch_size should be an integer'
        assert batch_size > 0, f'batch_size should be  > 0'
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.method = method

    def get_iterator(self, Data, Labels):
        """
        Creates a generator to iterate through the Data

        Args:
            :Data (numpy.ndarray): Input to be iterated over
            :Labels (numpy.ndarray): Labesl to be iterated over

        Returns:
            :generator (generator): return a generator that
                can be used to iterate over Data and Labels
        """
        n_train = Data.shape[0]
        assert n_train == Labels.shape[0], 'len(Data) not same as len(Labels)'
        if self.method == 'full_batch':
            self.batch_size = n_train
        if self.shuffle:
            order = np.random.permutation(n_train)
        else:
            order = np.arange(n_train)
        start_idx = 0
        while start_idx < n_train:
            end_idx = min(start_idx+self.batch_size, n_train)
            idxs = order[start_idx:end_idx]
            mb_inputs = Data[idxs]
            mb_labels = Labels[idxs]
            yield mb_inputs, mb_labels
            start_idx += self.batch_size

    def __repr__(self):
        return f'{self.method} iterator with shuffling = {self.shuffle}'


class minibatch_iterator(Iterator):
    '''
    mini_batch iterator

    Args:
        :batch_size (int): batch size to used
        :shuffle (bool): shuffles each time if set to True
    '''

    def __init__(self, batch_size=128, *, shuffle=True):
        '''
        Initializes the Iterator
        Args:
            batch_size (int) : batch size to used
            shuffle (bool)   : shuffles each time if set to True
        '''
        super(minibatch_iterator, self).__init__(batch_size=batch_size,
                                                 shuffle=shuffle,
                                                 method='mini_batch')

    def __repr__(self):
        return (f'mini_batch iterator '
                f'with batch_size = {self.batch_size} '
                f'and shuffling = {self.shuffle} ')


class fullbatch_iterator(Iterator):
    '''
    Full batch iterator

    Args:
        :shuffle (bool): shuffles each time if set to True
    '''

    def __init__(self, *, shuffle=True):
        '''
        Initializes the Iterator
        Args:
            shuffle (bool)   : shuffles each time if set to True
        '''
        super(fullbatch_iterator, self).__init__(batch_size=-1,
                                                 shuffle=shuffle,
                                                 method='full_batch')


class stochastic_iterator(Iterator):
    '''
    Stochastic iterator

    Args:
        :shuffle (bool): shuffles each time if set to True
    '''

    def __init__(self, *, shuffle=True):
        '''
        Initializes the Iterator
        Args:
            shuffle (bool): shuffles each time if set to True
        '''
        super(stochastic_iterator, self).__init__(batch_size=1,
                                                  shuffle=shuffle,
                                                  method='stochastic')



import numpy as np
from simflow.layers.layer_class import Layer


class Dense(Layer):
    '''Dense / Linear Layer

    Represent a linear transformation Y = X*W + b
        - X: is an numpy.ndarray with shape (batch_size, input_dim)
        - W: is a trainable matrix with dimensions (input_dim, output_dim)
        - b: is a bias with dimensions (1, output_dim)
        - Y: is an numpy.ndarray with shape (batch_size, output_dim)
    '''

    def __init__(self, input_dim, output_dim, *,
                 init_method='Xavier', trainable=True):
        '''
        Initializes the Desnse layer parameter
            - W: is initialized with either Xavier or He initialization
            - b: is initialized to zero

        Args:
            input_dim (int): size of input passed
            output_dim (int): size of output requred
            init_method (str): initialization method to be used for Weights
            trainable (bool): False parameters of the layer are frozen
                              True parameters updated during optimizer step
        '''
        self.init_method = init_method
        self.W = np.random.randn(input_dim, output_dim)
        self.W *= self._initializer_(self.W, init_method)
        self.b = np.zeros((1, output_dim))
        self.cache_in = None
        self.trainable = trainable
        self.l_name = 'Dense'
        self.params = [self.W, self.b]

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the Dense Layer

        Args:
            X (numpy.ndarray): Input array (batch_size x input_dim)
            train (bool): True -> enables gardient caching for backward step

        Returns:
            Out (numpy.ndarray): Output after transformation Y = X*W + b
                                 shape of output is (batch_size x output_dim)
        '''
        assert len(X.shape) == 2, "input dimenstions not supported"
        assert X.shape[1] == self.W.shape[0], (
            f"input dim doesn't match, each X has dimension {X.shape[1]} "
            f"but Weights defined are of shape {self.W.shape[0]}"
        )
        out = X@self.W + self.b
        if train:
            self.cache_in = X
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the Dense Layer

        Args:
            dY (numpy.ndarray): Output gradient backpropagated
                                shape of dY is (batch_size x output_dim)

        Returns:
            dX (numpy.ndarray): Input gradient
            var_grad_list (list):
                trainable = True: [(W,dW),(b,db)]
                trainable = False: [ ]
        '''
        dX = dY@self.W.T
        if self.trainable:
            if self.cache_in is None:
                raise RuntimeError(
                    'Gradient cache not defined. When training the train '
                    f'argument must be set to true in the forward pass.'
                )
            X = self.cache_in
            db = np.sum(dY, axis=0, keepdims=True)
            dW = X.T@dY
            assert X.shape == dX.shape, (
                f"Dimensions of grad and variable should match, "
                f"X has shape {X.shape} and dX has shape {dX.shape}"
            )
            assert self.W.shape == dW.shape, (
                f"Dimensions of grad and variable should match, "
                f"W has shape {self.W.shape} and dW has shape {dW.shape}"
            )
            assert self.b.shape == db.shape, (
                f"Dimensions of grad and variable should match, "
                f"b has shape {self.b.shape} and db has shape {db.shape}"
            )
            return dX, [(self.W, dW), (self.b, db)]
        # if not trainable
        return dX, []

    def __repr__(self):
        return f'Dense Layer with shape {self.W.shape}'

    def _get_config_(self):
        '''
        returns the dict of params required to recreate the layer
        '''
        input_dim, output_dim = self.W.shape
        config = {"input_dim": input_dim,
                  "output_dim": output_dim,
                  "init_method": self.init_method,
                  'trainable': self.trainable}
        base_config = super(Dense, self)._get_config_()
        return dict(list(base_config.items())+list(config.items()))


# adding aliases
Linear = Dense


class BN_mean(Layer):
    '''
    Mean only Batch normalization  Layer

    During Train:
        BN(x) = X - mean(X_batch) + beta

    During Test:
        BN(x) = X - mean_learned + beta
    '''

    def __init__(self, dim, *, elr=0.9, trainable=True):
        '''
        Initializes the BN_mean layer parameter
            beta is initialized to zero
            mean_learned is initialized to zero

        Args:
            dim (int): size of input passed
            elr (float): exponential learning rate for updating mean_learned
            trainable (bool): False parameters of the layer are frozen
                              True parameters are updated during optimizer step
        '''
        assert isinstance(elr, float), f'should be float'
        assert (0 < elr < 1), f'should be between 0 and 1 but given {elr}'
        self.dim = dim
        self.beta = np.zeros((1, int(np.prod(dim))))
        self.cache_in = None
        self.mean_learned = np.zeros_like(self.beta)
        self.elr = elr
        self.trainable = trainable
        self.l_name = 'Mean only Batchnorm'
        self.params = [self.beta]

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the BN_mean Layer

        Args:
            X (numpy.ndarray): Input array
            train (bool): Set to True to enable caching for backward step

        Returns:
            Out (numpy.ndarray): Output after applying BN_mean()
        '''
        X_shape = X.shape
        X_flat = X.reshape(X_shape[0], -1)
        if train:
            current_mean = np.mean(X_flat, axis=0, keepdims=True)
            out_flat = X_flat - current_mean + self.beta

            # update for mean_learned (exponential moving average no bias
            # correction since we are going to be trainig it sufficiently)
            self.mean_learned = (self.elr*self.mean_learned
                                 + current_mean*(1-self.elr))

        else:  # during test use mean_learned
            out_flat = X_flat - self.mean_learned + self.beta
        return out_flat.reshape(X_shape)

    def backward(self, dY):
        '''
        Performs a backward pass through the BN_mean Layer

        Args:
            dY (numpy.ndarray): Output gradient

        Returns:
            dX (numpy.ndarray): Input gradient
            var_grad_list (list):
                trainable = True: [(beta,dbeta)]
                trainable = False: [ ]
        '''
        dY_shape = dY.shape
        dY_flat = dY.reshape(dY_shape[0], -1)
        N, D = dY_flat.shape
        dx1 = dY_flat
        dx2 = np.ones((N, D))/N * -1 * np.sum(dY_flat, axis=0)
        dX_flat = dx1 + dx2
        dX = dX_flat.reshape(dY_shape)
        if self.trainable:
            dbeta = np.sum(dY_flat, axis=0, keepdims=True)
            return dX, [(self.beta, dbeta)]
        # if not trainable
        return dX, []

    def _get_config_(self):
        '''
        returns the dict of params required to recreate the layer
        '''
        config = {"dim": self.dim, "elr": self.elr,
                  "trainable": self.trainable}
        base_config = super(BN_mean, self)._get_config_()
        return dict(list(base_config.items())+list(config.items()))


class BN(Layer):
    '''
    Batch normalization  Layer (Full)

    During Train:
        BN(x) = gamma(X - mean(X_batch))/std(X_batch) + beta

    During Test:
        BN(x) = gamma((X - mean_learned)/Learned_std) + beta
    '''

    def __init__(self, dim, *, elr=0.9, trainable=True):
        '''
        Initializes the BN layer parameter
            beta is initialized to zeros
            gamma is initialized to zeros
            mean_learned is initialized to zeros
            var_learned is initialized to zeros

        Args:
            dim (int) : size of input passed
            elr (float) : exponential lr for updating mean_learned
            trainable (bool) : False -> layer is frozen
                               True -> updated during optimizer step

        '''
        assert isinstance(elr, float), f'should be float'
        assert (0 < elr < 1), f'should be between 0 and 1 but given {elr}'
        self.dim = dim
        self.beta = np.zeros((1, int(np.prod(dim))))
        self.gamma = np.zeros((1, int(np.prod(dim))))
        self.cache_in = None
        self.mean_learned = np.zeros_like(self.beta)
        self.var_learned = np.zeros_like(self.gamma)
        self.elr = elr
        self.trainable = trainable
        self.l_name = 'Batchnorm'
        self.eps = 1e-10  # to avoid division_by_zero error if var = 0
        self.params = [self.gamma, self.beta]

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the BN Layer

        Args:
            X (numpy.ndarray): Input array
            train (bool): Set to True to enable caching for backward step

        Returns:
            Out (numpy.ndarray): Output after applying BN() transformation
        '''
        X_shape = X.shape
        X_flat = X.reshape(X_shape[0], -1)
        if train:
            assert X_shape[0] > 1, "not supported in training mode"
            current_mean = np.mean(X_flat, axis=0)
            current_var = np.var(X_flat, axis=0)
            X_norm_flat = (X_flat - current_mean)/np.sqrt(current_var+self.eps)
            out_flat = self.gamma*X_norm_flat + self.beta

            # update for mean_learned and std_learned
            # (exponential moving average no bias currection
            # since we are going to be trainig it sufficiently)
            self.mean_learned = (self.elr*self.mean_learned
                                 + current_mean*(1-self.elr))
            self.var_learned = (self.elr*self.var_learned
                                + current_var*(1-self.elr))

            self.cache_in = (X_flat, current_mean, current_var, X_norm_flat)

        else:  # during test use mean_learned adn std_learned
            X_norm_flat = ((X_flat - self.mean_learned)
                           / np.sqrt(self.var_learned+self.eps))
            out_flat = self.gamma*X_norm_flat + self.beta
        return out_flat.reshape(X_shape)

    def backward(self, dY):
        '''
        Performs a backward pass through the BN Layer

        Args:
            dY (numpy.ndarray): Output gradient

        Returns:
            dX (numpy.ndarray): Input gradient after backpropagation
            var_grad_list (list):
                trainable = True: [(gamma,dgamma), (beta,dbeta)]
                trainable = False: []
        '''
        if self.cache_in is None:
            raise RuntimeError(
                f'Gradient cache not defined. When training the train '
                f'argument must be set to true in the forward pass.'
            )
        X_flat, current_mean, current_var, X_norm_flat = self.cache_in
        dY_shape = dY.shape
        # fatten dY
        dY_flat = dY.reshape(dY_shape[0], -1)
        N = dY_shape[0]
        X_mu = X_flat - current_mean
        inv_var = 1/np.sqrt(current_var+self.eps)

        dX_norm = dY_flat * self.gamma

        d_var = (np.sum(dX_norm*X_mu, axis=0)
                 * (-((current_var+self.eps)**(-3/2))/2))

        d_mu = (np.sum(dX_norm*(-inv_var), axis=0)
                + (1/N)*d_var*np.sum(-2*X_mu, axis=0))

        dX_flat = (dX_norm*inv_var)+(d_mu+2*d_var*X_mu)/N
        dX = dX_flat.reshape(dY_shape)
        if self.trainable:
            dbeta = np.sum(dY_flat, axis=0, keepdims=True)
            dgamma = np.sum(dY_flat*X_norm_flat, axis=0, keepdims=True)
            return dX, [(self.gamma, dgamma), (self.beta, dbeta)]
        # if not trainable
        return dX, []

    def _get_config_(self):
        '''
        returns the dict of params required to recreate the layer
        '''
        config = {"dim": self.dim, "elr": self.elr,
                  "trainable": self.trainable}
        base_config = super(BN, self)._get_config_()
        return dict(list(base_config.items())+list(config.items()))


class Flatten(Layer):
    '''
    Flatten layer
    takes a tensor and converts it to a matrix
    This layer usually acts as an interface between conv layer and dense layer
    '''

    def __init__(self):
        '''
        Initialization :
            Does nothing since nothing to initialize
        '''
        self.cache_in = None
        self.l_name = 'Flatten'
        self.params = []

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the Flatten Layer

        Args:
            :X (numpy.ndarray): Input array
            :train (bool): No effect of this layer

        Returns:
            :Out (numpy.ndarray): Output after flattening
        '''
        self.shape = X.shape
        out = X.reshape(self.shape[0], -1)
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the BN_mean Layer

        Args:
            :dY (numpy.ndarray): Output gradient

        Returns:
            :dX (numpy.ndarray): Input gradient after reshaping dY
            :var_grad_list (list): [], since layer is not trainable
        '''
        dX = dY.reshape(self.shape)
        return dX, []



import numpy as np
from simflow.layers.convolutional import dilated_Conv2D
from simflow.utils.grad_check_utils import numerical_gradient_array
import unittest


class TestDilatedConv(unittest.TestCase):
    def test_dilated_conv_layer_back_prop(self):
        eps = 1e-7
        batch_size = 32
        filter_size = 3
        h_x, w_x = 7, 7
        inChannels = 3
        n_filter = 5
        padding = 0
        stride = 1
        dilation = 2
        new_filter_size = dilation*(filter_size-1)+1
        h_out = (h_x - new_filter_size + 2*padding)//stride + 1
        w_out = (w_x - new_filter_size + 2*padding)//stride + 1

        x = np.random.randn(batch_size, inChannels, h_x, w_x)
        w = np.random.randn(n_filter, inChannels, filter_size, filter_size)
        b = np.random.randn(n_filter, 1)
        dout = np.random.randn(batch_size, n_filter, h_out, w_out)

        dc_layer = dilated_Conv2D(inChannels=inChannels, outChannels=n_filter,
                                  filter_size=filter_size, stride=stride,
                                  padding=padding, trainable=True)
        dc_layer.W = w
        dc_layer.b = b

        dx_num = numerical_gradient_array(dc_layer.forward,
                                          x, dout, h=eps)
        dw_num = numerical_gradient_array(lambda w: dc_layer.forward(x),
                                          w, dout, h=eps)
        db_num = numerical_gradient_array(lambda b: dc_layer.forward(x),
                                          b, dout, h=eps)

        dx, grads = dc_layer.backward(dout)
        dw, db = grads[0][1], grads[1][1]

        assert np.allclose(dx, dx_num, atol=eps)
        assert np.allclose(dw, dw_num, atol=eps)
        assert np.allclose(db, db_num, atol=eps)



import numpy as np
from abc import ABC, abstractmethod


class Optimizer(ABC):
    '''
    Optimizer Class:
    Used to update parameters

    **note**
    this is the parent class of all optimizers, not an actual optimizer
    that can be used for training models.

    '''

    def __init__(self, **kwargs):
        '''
        Initializes updates and weights to empty list

        Args:
            clipvalue (float): value to clip gradients to
        '''
        allowed_kwargs = {'clipvalue'}
        for k in kwargs:
            if k not in allowed_kwargs:
                raise TypeError('Unexpected keyword argument '
                                'passed to optimizer: ' + str(k))
        self.__dict__.update(kwargs)
        self.updates = []
        self.weights = []

    def get_var_and_grads(self, vars_and_grads):
        '''
        splits vars_and_grads into variables and gradients
        also clips gradients to clipvalue chosen before

        Args:
            vars_and_grads (List[Tuple[np.ndarray]]): variables and gradients
        Returns:
            params (List[np.ndarray]): pointers to variable to be updated
            grads (List[np.ndarray]): pointers to gradients to be updated
        '''
        try:
            params, grads = zip(*vars_and_grads)
        except ValueError:
            raise ValueError('no gradients found please re-check')

        if hasattr(self, 'clipvalue') and self.clipvalue > 0:
            grads = [np.clip(g, -self.clipvalue, self.clipvalue)
                     for g in grads]
        return params, grads

    @abstractmethod
    def update_step(self, vars_and_grads):
        '''abstractmethod update_step

        updates vara and grads
        '''


class SGD(Optimizer):
    '''Stochastic Gradient Descent

    Stochastic Gradient Descent combined method of the following methods of
    gradient Descent

    Vanilla:
        Takes a step proportional to the negative of the gradient
            W(t) = W(t-1) - lr * grad

    Momentum:
        Keeps track of velocity and takes a step in the direction of velocity
            v(t) = momentum*v(t-1) - lr*grad
            W(t) = W(t-1) + v(t)

    Nestrov Momentum:
        Smarter momentum step in direction of velocity, correct based on grad
            v(t) = momentum*v(t-1) - lr*grad
            W(t) = W(t-1) -lr*grad + momentum*v(t)
    '''

    def __init__(self, lr=0.01, momentum=0, decay=0, nestrov=False, **kwargs):
        """Setup SGD

        Args:
            lr(float): learning rate [default = 0.01]
            momentum (float): momentum factor used [default = 0]
            decay(float): decay factor by which learning rate
                reduces [default = 0]
            nestrov (bool): set True to enable Nestrov  [default = False]
            clipvalue (float): value to clip gradients to [default = inf]
        """
        super(SGD, self).__init__(**kwargs)
        assert isinstance(lr, float)
        assert isinstance(momentum, (float, int))
        assert isinstance(decay, (float, int))
        assert isinstance(nestrov, bool)
        assert decay >= 0, "-ve decay not valid"
        assert momentum >= 0, "-ve momentum not valid"
        assert momentum < 1, f"momentum should be <1, currently {momentum}"
        assert lr > 0, f"lr should be >0,currently {lr}"
        self.lr = lr
        self.momentum = momentum
        self.decay = decay
        self.nestrov = nestrov
        self.iter = 0

    def update_step(self, vars_and_grads):
        '''
        updates vara and grads using SGD

        Args:
            vars_and_grads (List[Tuple[np.ndarray]]) : variables and gradients
        '''
        params, grads = self.get_var_and_grads(vars_and_grads)
        if not hasattr(self, 'v_grads') and self.momentum:
            # for the first time we need  to init v_grads with zeros
            self.v_grads = [np.zeros_like(g) for g in grads]
        self.iter += 1
        lr = self.lr
        lr *= (1/(1+self.decay*self.iter))

        if self.momentum:
            for p, g, v in zip(params, grads, self.v_grads):
                v *= self.momentum
                v -= (lr * g)
                if self.nestrov:
                    p += self.momentum*v - lr*g
                else:
                    p += v
        else:
            for p, g in zip(params, grads):
                p -= lr*g


class RMSProp(Optimizer):
    '''RMSProp

    Gradient Descent using RMSProp algorithm

    References:
        - [rmsprop: Divide the gradient by a running average
           ](http://www.cs.toronto.edu/~tijmen/csc321/slides/lecture_slides_lec6.pdf)
    '''

    def __init__(self, lr=0.001, rho=0.9, decay=0, eps=1e-7, **kwargs):
        """Setup RMSProp

        Args:
            lr(float): learning rate [default = 0.001]
            rho (float): RMSProp decay factor used [default = 0.9]
            decay(float): decay factor [default = 0]
            eps (float): Fuzziness factor  [default = 1e-7]
            clipvalue (float): value to clip gradients to [default = inf]
        """
        super(RMSProp, self).__init__(**kwargs)
        assert isinstance(lr, float)
        assert isinstance(rho, (float, int))
        assert isinstance(decay, (float, int))
        assert isinstance(eps, (float, int))

        assert decay >= 0, "-ve decay not valid"
        assert rho >= 0, "-ve rho not valid"
        assert rho < 1, f"rho should be <1, currently {rho}"
        assert lr > 0, f"lr should be >0,currently {lr}"
        self.lr = lr
        self.rho = rho
        self.decay = decay
        self.iter = 0
        self.eps = eps
        # precaution to avoid divide by zero
        if self.eps == 0:
            self.eps = 1e-7

    def update_step(self, vars_and_grads):
        '''
        updates vara and grads using RMSProp

        Args:
            vars_and_grads (List[Tuple[np.ndarray]]): variables and gradients
        '''
        params, grads = self.get_var_and_grads(vars_and_grads)
        if not hasattr(self, 'a_grads'):
            # for the first time we need  to init a_grads with zeros
            self.a_grads = [np.zeros_like(g) for g in grads]

        self.iter += 1
        lr = self.lr
        lr *= (1/(1+self.decay*self.iter))

        for p, g, a in zip(params, grads, self.a_grads):
            a *= self.rho
            a += (1-self.rho) * np.square(g)
            p -= lr * (g/(np.sqrt(a)+self.eps))


class Adagrad(Optimizer):
    '''
    Adagrad Gradient Descent

    Adagrad is an optimizer with parameter-specific learning rates,
    which are adapted relative to how frequently a parameter gets
    updated during training. The more updates a parameter receives,
    the smaller the updates.

    Suffers from the inherent problem of early stopping

    References:
        - [Adaptive Subgradient Methods for Online Learning and Stochastic
           Optimization](http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf)
    '''

    def __init__(self, lr=0.01, decay=0, eps=1e-7, **kwargs):
        """Setup Adagrad

        Args:
            lr(float): learning rate [default = 0.01]
            decay(float): decay factor [default = 0]
            eps (float): Fuzziness factor  [default = 1e-7]
            clipvalue (float): value to clip gradients to [default = inf]
        """
        super(Adagrad, self).__init__(**kwargs)
        assert isinstance(lr, float)
        assert isinstance(decay, (float, int))
        assert isinstance(eps, (float, int))

        assert decay >= 0, "-ve decay not valid"
        assert lr > 0, f"lr should be >0,currently {lr}"
        self.lr = lr
        self.decay = decay
        self.iter = 0
        self.eps = eps
        if self.eps == 0:
            self.eps = 1e-7

    def update_step(self, vars_and_grads):
        '''
        updates vara and grads using Adagrad

        Args:
            vars_and_grads (List[Tuple[np.ndarray]]): variables and gradients
        '''
        params, grads = self.get_var_and_grads(vars_and_grads)
        if not hasattr(self, 'a_grads'):
            # for the first time we need  to init a_grads with zeros
            self.a_grads = [np.zeros_like(g) for g in grads]
        self.iter += 1
        lr = self.lr
        lr *= (1/(1+self.decay*self.iter))

        for p, g, a in zip(params, grads, self.a_grads):
            a += np.square(g)
            p -= lr * (g/(np.sqrt(a)+self.eps))


class Adadelta(Optimizer):
    '''
    Adadelta Gradient Descent

    Adadelta is an optimizer with parameter-specific learning rates,
    which are adapted relative to how frequently a parameter gets
    updated during training. The more updates a parameter receives,
    the smaller the updates.

    But overcomes the inherent problem of early stopping in Adadelta,
    since it accumulates only gradients within a fixed window.

    References:
        - [Adadelta - an adaptive learning rate method](
           https://arxiv.org/abs/1212.5701)
    '''

    def __init__(self, lr=1, rho=0.95, decay=0, eps=1e-7, **kwargs):
        """

        Args:
            lr(float): learning rate [default = 0.001]
            rho(float): Adadelta decay factor
            decay(float): decay factor [default = 0]
            eps (float): Fuzziness factor  [default = 1e-7]
            clipvalue (float): value to clip gradients to [default = inf]
        """
        super(Adadelta, self).__init__(**kwargs)
        assert isinstance(lr, (float, int))
        assert isinstance(decay, (float, int))
        assert isinstance(rho, (float, int))
        assert isinstance(eps, (float, int))

        assert decay >= 0, "-ve decay not valid"
        assert lr > 0, f"lr should be > 0,currently {lr}"
        assert rho >= 0, f"rho should be >= 0,currently {rho}"
        assert rho < 1, f"rho should be < 1, currently {rho}"

        self.lr = lr
        self.rho = rho
        self.decay = decay
        self.iter = 0
        self.eps = eps
        if self.eps == 0:
            self.eps = 1e-7

    def update_step(self, vars_and_grads):
        '''
        updates vara and grads using Adagrad

        Args:
            vars_and_grads (List[Tuple[np.ndarray]]) : variables and gradients
        '''
        params, grads = self.get_var_and_grads(vars_and_grads)
        if not hasattr(self, 'a_grads'):
            # for the first time we need  to init a_grads with zeros
            self.a_grads = [np.zeros_like(g) for g in grads]
        if not hasattr(self, 'da_grads'):
            # for the first time we need  to init da_grads with zeros
            self.da_grads = [np.zeros_like(g) for g in grads]

        self.iter += 1

        lr = self.lr
        lr *= (1/(1+self.decay*self.iter))

        for p, g, a, d_a in zip(params, grads, self.a_grads, self.da_grads):
            a *= self.rho
            a += (1-self.rho) * np.square(g)

            update = g * np.sqrt(d_a+self.eps) / np.sqrt(a+self.eps)

            p -= lr * update

            d_a *= self.rho
            d_a += (1-self.rho)*np.square(update)


class Adam(Optimizer):
    '''
    Adam
    Adaptive Moment Estimation in short combination of momentum and RMSProp

    References:
        - [Adam - A Method for Stochastic Optimization](
           https://arxiv.org/abs/1412.6980v8)
        - [On the Convergence of Adam and Beyond](
           https://openreview.net/forum?id=ryQu7f-RZ)
    '''

    def __init__(self, lr=0.001, beta_1=0.9, beta_2=0.999, decay=0,
                 eps=0, amsgrad=False, curve_correction=False, **kwargs):
        """Setup Adam

        Args:
            lr(float): learning rate [default = 0.001]
            beta_1 (float): first moment factor used [default = 0.9]
            beta_2 (float): second moment factor used [default = 0.999]
            decay(float): decay rate [default = 0]
            amsgrad (bool): set True to enable AMSGrad  [default = False]
            curve_correction (bool): set to True to enable
                curve_correction of grads [defaul = False]
            clipvalue (float): value to clip gradients to [default = inf]
        """
        super(Adam, self).__init__(**kwargs)
        assert isinstance(lr, float)
        assert isinstance(beta_1, (float, int))
        assert isinstance(beta_2, (float, int))

        assert isinstance(decay, (float, int))
        assert isinstance(amsgrad, bool)
        assert isinstance(curve_correction, bool)

        assert decay >= 0, "-ve decay not valid"
        assert beta_1 >= 0 and beta_2 >= 0, "-ve moments not valid"
        assert beta_1 < 1 and beta_2 < 1, f"both moments should be <1"
        assert lr > 0, f"lr should be >0,currently {lr}"
        self.lr = lr
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.decay = decay
        self.amsgrad = amsgrad
        self.iter = 0
        self.eps = eps
        self.curve_correction = curve_correction
        if self.eps == 0:
            self.eps = 1e-7

    def update_step(self, vars_and_grads):
        '''
        updates vara and grads using SGD

        Args:
            vars_and_grads (List[Tuple[np.ndarray]]) : variables and gradients
        '''
        params, grads = self.get_var_and_grads(vars_and_grads)
        if not hasattr(self, 'm_grads'):
            # for the first time we need  to init m_grads with zeros
            self.m_grads = [np.zeros_like(g) for g in grads]
        if not hasattr(self, 'v_grads'):
            # for the first time we need  to init v_grads with zeros
            self.v_grads = [np.zeros_like(g) for g in grads]

        if not hasattr(self, 'vhats'):
            if self.amsgrad:
                self.vhats = [np.zeros_like(g) for g in grads]
            else:
                self.vhats = [np.zeros(1) for _ in grads]

        self.iter += 1

        lr = self.lr
        lr *= (1/(1+self.decay*self.iter))

        for p, g, m, v, vhat in zip(params, grads, self.m_grads,
                                    self.v_grads, self.vhats):
            # update first moment
            m *= self.beta_1
            m += (1-self.beta_1) * g

            # curve correction
            if self.curve_correction:
                m_c = m / (1-self.beta_1**self.iter)
            else:
                m_c = m

            # update second moment
            v *= self.beta_2
            v += (1-self.beta_2) * np.square(g)
            # curve_correction
            if self.curve_correction:
                v_c = v / (1-self.beta_2**self.iter)
            else:
                v_c = v

            if self.amsgrad:
                vhat = np.maximum(vhat, v_c)
                p -= lr * m_c / (np.sqrt(vhat)+self.eps)
            else:
                p -= lr * m_c / (np.sqrt(v_c)+self.eps)



import numpy as np


def load_normalized_mnist_data_flat():
    '''
    Loads and normalizes the MNIST data. Reads the data from

        -data/mnist_train.csv
        -data/mnist_test.csv

    If these are not present they are downloaded and saved from
    https://pjreddie.com/projects/mnist-in-csv/

    Returns:
        :input (dict of numpy.ndarray):  Input
        :labels (dict of numpy.ndarray):  Lables

    Each has keys 'train', 'val', 'test' which map to numpy arrays
    '''
    try:
        print('loading data please wait will take around a minute')
        data = np.loadtxt('data/mnist_train.csv', dtype=int, delimiter=',')
    except OSError:
        import os
        import wget
        url = 'https://pjreddie.com/media/files/mnist_train.csv'
        print('File now found downloading from online source')
        print('Downloading data will take some time')
        try:
            os.mkdir('data')
        except FileExistsError:
            pass
        wget.download(url, 'data/mnist_train.csv')
        print('\nNow loading data please wait will take around a minute')

        data = np.loadtxt('data/mnist_train.csv', dtype=int, delimiter=',')

    try:
        print('loading data please wait will take around a minute')
        test_data = np.loadtxt('data/mnist_test.csv', dtype=int, delimiter=',')
    except OSError:
        import os
        import wget
        url = 'https://pjreddie.com/media/files/mnist_test.csv'
        print('File now found downloading from online source')
        print('Downloading data will take some time')
        try:
            os.mkdir('data')
        except FileExistsError:
            pass
        wget.download(url, 'data/mnist_test.csv')
        print('\nNow loading data please wait will take around a minute')

        test_data = np.loadtxt('data/mnist_test.csv', dtype=int, delimiter=',')

    inputs = dict()
    labels = dict()

    train_data = data[:50000]
    train_inputs = train_data[:, 1:]

    val_data = data[50000:]
    val_inputs = val_data[:, 1:]

    test_inputs = test_data[:, 1:]

    mean = np.mean(train_inputs)
    std = np.std(train_inputs)

    inputs['train'] = (train_inputs - mean)/std
    inputs['val'] = (val_inputs - mean)/std
    inputs['test'] = (test_inputs - mean)/std

    labels['train'] = train_data[:, 0]
    labels['val'] = val_data[:, 0]
    labels['test'] = test_data[:, 0]
    print('Data loading completed')
    return inputs, labels


def load_normalized_mnist_data_conv():
    '''
    Loads and normalizes the MNIST data. Reads the data from

        -data/mnist_train.csv
        -data/mnist_test.csv

    If these are not present they are downloaded and saved from
    https://pjreddie.com/projects/mnist-in-csv/

    Returns:
        :input (dict of numpy.ndarray):  Input
        :labels (dict of numpy.ndarray):  Lables

    Each has keys 'train', 'val', 'test' which map to numpy arrays
    '''
    print('loading data please wait will take around a minute')
    data = np.loadtxt('data/mnist_train.csv', dtype=int, delimiter=',')
    test_data = np.loadtxt('data/mnist_test.csv', dtype=int, delimiter=',')
    inputs = dict()
    labels = dict()

    train_data = data[:50000]
    train_inputs = train_data[:, 1:].reshape(-1, 1, 28, 28)

    val_data = data[50000:]
    val_inputs = val_data[:, 1:].reshape(-1, 1, 28, 28)

    test_inputs = test_data[:, 1:].reshape(-1, 1, 28, 28)

    mean = np.mean(train_inputs)
    std = np.std(train_inputs)

    inputs['train'] = (train_inputs - mean)/std
    inputs['val'] = (val_inputs - mean)/std
    inputs['test'] = (test_inputs - mean)/std

    labels['train'] = train_data[:, 0]
    labels['val'] = val_data[:, 0]
    labels['test'] = test_data[:, 0]
    print('Data Loading completed')
    return inputs, labels



import numpy as np
from abc import ABC, abstractmethod


class Loss(ABC):
    '''
    Abstract class representing a loss function
    '''
    @abstractmethod
    def get_loss(self, scores, labels):
        """
        will return the loss value
        """


class SoftmaxCrossEntropyLoss(Loss):
    '''
    Represents the categorical softmax cross entropy loss
    '''

    def get_loss(self, scores, labels):
        '''
        Calculates the average categorical softmax cross entropy loss.

        Args:
            scores (numpy.ndarray): Unnormalized logit class scores.
            labels (numpy.ndarray): True labels represented as ints
        Returns:
            loss (float): The average cross entropy
            grad (numpy.ndarray): Gradient for scores w.r.t the loss.
        '''
        scores_norm = scores - np.max(scores, axis=1, keepdims=True)
        scores_norm = np.exp(scores_norm)
        scores_norm = scores_norm / np.sum(scores_norm, axis=1, keepdims=True)

        true_class_scores = scores_norm[np.arange(len(labels)), labels]
        loss = np.mean(-np.log(true_class_scores))

        one_hot = np.zeros(scores.shape)
        one_hot[np.arange(len(labels)), labels] = 1.0
        grad = (scores_norm - one_hot) / len(labels)

        return loss, grad



import numpy as np
from simflow.layers.activations import (
    ReLU, LeakyReLU, Softplus, exp, Sigmoid, Tanh
)
from simflow.utils.grad_check_utils import numerical_gradient_array
import unittest


class TestActivations(unittest.TestCase):
    def setUp(self):
        self.eps = 1e-7
        self.batch_size = 32
        self.h_x, self.w_x = 7, 7
        self.inChannels = 3

        self.x = np.random.randn(self.batch_size, self.inChannels,
                                 self.h_x, self.w_x)
        self.dout = np.random.randn(self.batch_size, self.inChannels,
                                    self.h_x, self.w_x)

    def helper_grad_check(self, Layer):
        layer = Layer(trainable=True)
        dx_num = numerical_gradient_array(lambda x: layer.forward(self.x),
                                          self.x, self.dout, h=self.eps)
        dx, _ = layer.backward(self.dout)

        assert np.allclose(dx, dx_num, atol=self.eps)

        with self.assertRaises(RuntimeError):
            layer2 = Layer(trainable=False)
            layer2.backward(self.dout)

    def test_relu_back_prop(self):
        self.helper_grad_check(ReLU)

    def test_LeakyReLU_back_prop(self):
        self.helper_grad_check(LeakyReLU)

    def test_Softplus_back_prop(self):
        self.helper_grad_check(Softplus)

    def test_exp_back_prop(self):
        self.helper_grad_check(exp)

    def test_sigmoid_back_prop(self):
        self.helper_grad_check(Sigmoid)

    def test_tanh_back_prop(self):
        self.helper_grad_check(Tanh)



import numpy as np

from simflow.optimizers import Optimizer
from simflow.iterators import Iterator
from simflow.layers import Layer
from simflow.losses import Loss


class Model(object):
    '''
    Represents a neural network with any combination of layers
    '''

    def __init__(self):
        '''
        Returns a new empty neural network with no layers or loss
        '''
        self.layers = []
        self.loss = None

    def summary(self):
        '''
        Prints a table describing the model layers and parameters used
        '''
        # currently not the most elegant implementation
        dash = "-"*75
        print(dash)
        print('{:<30s}|{:22s}|{:20s}'.format("Name", "Trainable Parameters",
                                             "Total Parameters"))
        print(dash)
        total_params = 0
        total_trainable_params = 0

        for layer in self.layers:
            param_size = 0
            trainable_param_size = 0
            for p in layer.params:
                param_size += np.prod(p.shape)
                if layer.trainable:
                    trainable_param_size += np.prod(p.shape)
            print('{:<30s}|{:>22d}|{:>20d}'.format(layer.l_name,
                                                   trainable_param_size,
                                                   param_size))
            total_params += param_size
            total_trainable_params += trainable_param_size
        print(dash)
        print('{:<30s}|{:>22d}|{:>20d}'.format("Total",
                                               total_trainable_params,
                                               total_params))
        print(dash)

    def get_params(self):
        '''
        Retruns the list of params in the model
        '''
        param_list = []
        for layer in self.layers:
            param_list += [(layer.l_name, layer.get_params())]
        return param_list

    def set_params(self, params):
        '''
        Sets the params of a layer with a new params

        Ags:
            :params (list of numpy.ndarray): new weights
        '''
        old_params = self.get_params()
        assert len(old_params) == len(params), 'Length missmatch'
        assert all((new_param[0] == old_param[0] and all(n.shape == o.shape)
                    for n, o in zip(new_param, old_param))
                   for (new_param, old_param)
                   in zip(params, old_params)), 'Structure missmatch'
        for (new_param, layer) in zip(params, self.layers):
            layer.set_params(new_param[1])

    def add_layer(self, layer):
        '''
        Adds a layer to the network in a sequential manner.
        The input to this layer will be the output of the last added layer
        or the initial inputs to the networks if this is the first layer added.

        Args:
            :layer (Layer): Layer to be added to the model
        '''
        assert isinstance(layer, Layer)
        self.layers.append(layer)

    def set_loss_fn(self, loss):
        '''
        Sets the loss fuction that the network uses for training

        Args:
            :loss (Loss): Loss function to be used
        '''
        assert isinstance(loss, Loss)
        self.loss = loss

    def predict(self, inputs, train=False):
        '''
        Calculates the output of the network for the given inputs.

        Args:
            :inputs (numpy.ndarray): Inputs to the network
        Returns:
            :output (numpy.ndarray): Outputs of the last layer of the network.
        '''
        output = inputs
        for layer in self.layers:
            output = layer.forward(output, train=train)
        return output

    def _forward_backward_(self, inputs, labels):
        '''
        Calculates the loss of the network for the given inputs and labels

        Args:
            :inputs (numpy.ndarray): Inputs to the network
            :labels (numpy.ndarray): Int representation of the labels
                (eg. the third class is represented by 2)
        Returns:
            :loss (float): The loss before updating the network
            :vars_and_grads (list of tuples): variables and their gradients
        '''
        vars_and_grads = []

        # Forward pass
        output = self.predict(inputs, train=True)

        # Backward pass
        loss, grad = self.loss.get_loss(output, labels)
        for layer in reversed(self.layers):
            grad, layer_var_grad = layer.backward(grad)
            vars_and_grads += layer_var_grad

        return loss, vars_and_grads

    def set_optimizer(self, optimizer):
        '''
        Sets the optmizier to be used

        Args:
            :optimizer (Optimizer): Optimizer to be used
        '''
        assert isinstance(optimizer, Optimizer)
        self.optimizer = optimizer

    def set_iterator(self, iterator):
        '''
        Sets up the iterator

        Args:
            :iterator (Iterator): Iterator to be used
        '''
        assert isinstance(iterator, Iterator)
        self.iterator = iterator

    def fit(self, Data, Labels, epochs=1, *, verbose=True, **kwargs):
        '''
        Trains the model on the provided Data and Labels

        Updates the model parameters which are trainable

        Args:
            :Data (numpy.ndarray): Data to fit on
            :Labels (numpy.ndarray): Labels for the Data
            :epochs (int): Number of epochs to train
            :verbose (bool): Set to True to print progress, default True
        Kwargs:
            :optimizer(optimizer): Optimizer to be used
            :iterator (Iterator): Iterator to be used
        '''
        allowed_kwargs = {'optimizer', 'iterator'}
        for k in kwargs:
            if k not in allowed_kwargs:
                raise TypeError('Unexpected keyword argument '
                                'passed to fit: ' + str(k))

            if k == 'optimizer':
                self.set_optimizer(kwargs[k])
            if k == 'iterator':
                self.set_iterator(kwargs[k])
        # self.__dict__.update(kwargs)
        if not hasattr(self, 'optimizer'):
            raise NameError('Optimizer not defined')
        if not hasattr(self, 'iterator'):
            raise NameError('Iterator not defined')
        for epoch in range(epochs):
            total_loss = 0
            for curr_Data, curr_Labels in self.iterator.get_iterator(Data,
                                                                     Labels):
                loss, vars_and_grads = self._forward_backward_(curr_Data,
                                                               curr_Labels)
                self.optimizer.update_step(vars_and_grads)
                total_loss += loss
            average_loss = total_loss/Data.shape[0]
            if verbose:
                prnt_tmplt = ('Epoch: {:3}, average train loss: {:0.3f}')
                print(prnt_tmplt.format(epoch, average_loss))

    def score(self, Data, Labels):
        '''
        Return loss and accuracy of a model on Data and Labels passed

        Args:
            :Data (numpy.ndarray): Data to find performance on
            :Labels (numpy.ndarray): Labels to find performance on
        '''
        assert hasattr(
            self, "loss"), 'please set a loss function to score with'
        assert Data.shape[0] == Labels.shape[0], 'should same'
        scores = self.predict(Data)
        loss, _ = self.loss.get_loss(scores, Labels)
        pred = np.argmax(scores, axis=1)
        correct = np.sum(pred == Labels)
        n_inp = Data.shape[0]
        avg_loss = loss/n_inp
        accuracy = correct/n_inp
        return avg_loss, accuracy



import numpy as np
from simflow.layers.layers import Dense
from simflow.utils.grad_check_utils import numerical_gradient_array
import unittest


class TestDense(unittest.TestCase):
    @staticmethod
    def standard_forward_Dense(X, w, b):
        return X@w + b

    def test_Dense_forward(self):
        batch_size = 32
        inp_dim = 10
        out_dim = 5
        eps = 1e-7
        Input = np.random.randn(batch_size, inp_dim)

        d_layer = Dense(inp_dim, out_dim, init_method='He')

        W = np.random.randn(inp_dim, out_dim)
        b = np.random.randn(1, out_dim)

        d_layer.W = W
        d_layer.b = b
        assert np.allclose(d_layer.forward(Input), self.standard_forward_Dense(
            Input, W, b), atol=eps), "forward not working"

    def test_linear_back(self):
        eps = 1e-7
        input_dim = 10
        output_dim = 5

        batch_size = 32

        random_inp = np.random.randn(batch_size, input_dim)
        W = np.random.randn(input_dim, output_dim)
        b = np.random.randn(1, output_dim)
        random_output_gradient = np.random.randn(batch_size, output_dim)

        lin_layer = Dense(input_dim, output_dim, init_method='He')
        lin_layer.W = W
        lin_layer.b = b

        dx_num = numerical_gradient_array(lambda x: lin_layer.forward(
            random_inp), random_inp, random_output_gradient, h=eps)
        dw_num = numerical_gradient_array(lambda W: lin_layer.forward(
            random_inp), W, random_output_gradient, h=eps)
        db_num = numerical_gradient_array(lambda b: lin_layer.forward(
            random_inp), b, random_output_gradient, h=eps)

        dx, grads = lin_layer.backward(random_output_gradient)
        dW, db = grads[0][1], grads[1][1]

        assert np.allclose(dx, dx_num, atol=eps)
        assert np.allclose(dW, dw_num, atol=eps)
        assert np.allclose(db, db_num, atol=eps)



# source https://github.com/yunjey/cs231n/tree/master/assignment2/cs231n
import numpy as np


def get_im2col_indices(x_shape, field_height=3,
                       field_width=3, padding=1, stride=1):
    # First figure out what the size of the output should be
    N, C, H, W = x_shape
    assert (H + 2 * padding - field_height) % stride == 0
    assert (W + 2 * padding - field_height) % stride == 0
    out_height = (H + 2 * padding - field_height) / stride + 1
    out_width = (W + 2 * padding - field_width) / stride + 1

    i0 = np.repeat(np.arange(field_height, dtype='int32'), field_width)
    i0 = np.tile(i0, C)
    i1 = stride * np.repeat(np.arange(out_height, dtype='int32'), out_width)
    j0 = np.tile(np.arange(field_width), field_height * C)
    j1 = stride * np.tile(np.arange(out_width, dtype='int32'), int(out_height))
    i = i0.reshape(-1, 1) + i1.reshape(1, -1)
    j = j0.reshape(-1, 1) + j1.reshape(1, -1)

    k = np.repeat(np.arange(C, dtype='int32'),
                  field_height * field_width).reshape(-1, 1)

    return (k, i, j)


def im2col_indices(x, field_height=3, field_width=3, padding=1, stride=1):
    # Zero-pad the input
    p = padding
    x_padded = np.pad(x, ((0, 0), (0, 0), (p, p), (p, p)), mode='constant')

    k, i, j = get_im2col_indices(x.shape, field_height, field_width, padding,
                                 stride)

    cols = x_padded[:, k, i, j]
    C = x.shape[1]
    cols = cols.transpose(1, 2, 0).reshape(field_height * field_width * C, -1)
    return cols


def col2im_indices(cols, x_shape, field_height=3, field_width=3, padding=1,
                   stride=1):
    N, C, H, W = x_shape
    H_padded, W_padded = H + 2 * padding, W + 2 * padding
    x_padded = np.zeros((N, C, H_padded, W_padded), dtype=cols.dtype)
    k, i, j = get_im2col_indices(x_shape, field_height, field_width, padding,
                                 stride)
    cols_reshaped = cols.reshape(C * field_height * field_width, -1, N)
    cols_reshaped = cols_reshaped.transpose(2, 0, 1)
    np.add.at(x_padded, (slice(None), k, i, j), cols_reshaped)
    if padding == 0:
        return x_padded
    return x_padded[:, :, padding:-padding, padding:-padding]



import numpy as np
from simflow.layers.layers import BN_mean
from simflow.utils.grad_check_utils import numerical_gradient_array
import unittest


class TestBatchNormMean(unittest.TestCase):
    def test_bn_mean_back_flat(self):
        eps = 1e-7
        dim = 128
        batch_size = 32

        random_inp = np.random.randn(batch_size, dim)
        beta = np.random.randn(1, dim)
        random_output_gradient = np.random.randn(batch_size, dim)
        learned_mean = np.random.randn(1, dim)

        b_layer = BN_mean(dim)
        b_layer.beta = beta
        b_layer.learned_mean = learned_mean

        dx_num = numerical_gradient_array(lambda x: b_layer.forward(
            random_inp), random_inp, random_output_gradient, h=eps)
        dbeta_num = numerical_gradient_array(lambda beta: b_layer.forward(
            random_inp), beta, random_output_gradient, h=eps)

        dx, grads = b_layer.backward(random_output_gradient)
        dbeta = grads[0][1]

        assert np.allclose(dx, dx_num, atol=eps)
        assert np.allclose(dbeta, dbeta_num, atol=eps)

    def test_bn_mean_back_3D(self):
        eps = 1e-7
        dim = (4, 5, 9)
        batch_size = 7

        random_inp = np.random.randn(batch_size, *dim)
        beta = np.random.randn(1, int(np.prod(dim)))
        random_output_gradient = np.random.randn(batch_size, *dim)
        learned_mean = np.random.randn(1, int(np.prod(dim)))

        b_layer = BN_mean(dim)
        b_layer.beta = beta
        b_layer.learned_mean = learned_mean

        dx_num = numerical_gradient_array(lambda x: b_layer.forward(
            random_inp), random_inp, random_output_gradient, h=eps)
        dbeta_num = numerical_gradient_array(lambda beta: b_layer.forward(
            random_inp), beta, random_output_gradient, h=eps)

        dx, grads = b_layer.backward(random_output_gradient)
        dbeta = grads[0][1]
        assert np.allclose(dx, dx_num, atol=eps)
        assert np.allclose(dbeta, dbeta_num, atol=eps)



import numpy as np

from simflow.utils import im2col
from simflow.layers.layer_class import Layer


class Conv2D(Layer):
    '''
    2D Convolutional Layer
    '''

    def __init__(self, outChannels, inChannels, filter_size, stride=1,
                 padding=0, *, init_method='Xavier', trainable=False):
        '''
        Initialization:
            W: initialized with either Xavier or He initialization
            b: initialized to zero

        Args:
            outChannels (int): Number of output channels
            inChannels (int): size of output requred
            filter_size (int): Size of each kernel (n x n)
            stride (int): Stride to be used
            padding (int): Padding to be used for convolution
            init_method (str): initialization method to be used for Weights
            trainable (bool):
                False: parameters of the layer are frozed
                True: parameters are updated during optimizer step
        '''
        assert isinstance(outChannels, int) and outChannels > 0
        assert isinstance(inChannels, int) and inChannels > 0
        assert isinstance(filter_size, int) and filter_size > 0
        assert isinstance(stride, int) and stride > 0
        assert isinstance(padding, int) and padding >= 0

        self.cache_in = None
        # currently supports only square kernels
        self.W_size = (outChannels, inChannels, filter_size, filter_size)
        self.W = np.random.rand(*self.W_size)
        self.W *= self._initializer_(self.W, init_method)
        self.b = np.zeros((outChannels, 1))
        self.stride = stride
        self.padding = padding
        self.trainable = trainable
        self.l_name = 'Conv2D'
        self.params = [self.W, self.b]

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the 2D Convolution layer:
        Convolves Inputs with the Weights

        Args:
            X (numpy.ndarray): Input array
            train (bool): Set true to cache X and X_col

        Returns:
            Out (numpy.ndarray): Output after Convolution
        '''
        output, X_col = self._convolve_(
            Input=X, kernel=self.W, bias=self.b,
            padding=self.padding, stride=self.stride
        )
        if train:
            self.cache_in = (X, X_col)
        return output

    def backward(self, dY):
        '''
        Performs a backward pass through the Conv2D Layer

        Args:
            dY (numpy.ndarray): Output gradient backpropagated from layers

        Returns:
            dX (numpy.ndarray): Input gradient after backprop dY through Conv2D
            var_grad_list (list):
                trainable = True [(W,dW), (b,db)]
                trainable = False [ ]
        '''
        if self.cache_in is None:
            raise RuntimeError('Gradient cache not defined')
        X, X_col = self.cache_in
        n_filter, d_filter, h_filter, w_filter = self.W.shape

        dY_reshaped = dY.transpose(1, 2, 3, 0).reshape(n_filter, -1)

        W_reshape = self.W.reshape(n_filter, -1)
        dX_col = W_reshape.T @ dY_reshaped
        dX = im2col.col2im_indices(dX_col, X.shape, h_filter, w_filter,
                                   padding=self.padding, stride=self.stride)
        assert X.shape == dX.shape, 'shape missmatch'
        if self.trainable:

            db = np.sum(dY, axis=(0, 2, 3)).reshape(n_filter, -1)

            dW = dY_reshaped @ X_col.T
            dW = dW.reshape(self.W.shape)

            assert self.W.shape == dW.shape, 'shape missmatch'
            assert self.b.shape == db.shape, 'shape missmatch'

            return dX, [(self.W, dW), (self.b, db)]
        # if not trainable
        return dX, []

    def __repr__(self):
        return (f'Conv2D Layer with {self.W.shape[0]}'
                f' number of filters of shape {self.W.Shape[1:]},'
                f' Stide = {self.stride}, padding = {self.padding}'
                f' Trainable = {self.trainable}')

    @staticmethod
    def _convolve_(Input, kernel, bias=0, padding=0, stride=1):
        '''
        2D Convolution function:
        Convolves Inputs with the given kernel

        Args:
            Input (numpy.ndarray): Input to be Convolved over
            kernel (numpy.ndarray): Kernel to be Convoled with
            bias (numpy.ndarray): bias optional, set to zero unless needed
            padding (int): padding to be used
            stride (int): stride to used

        Returns:
            out (numpy.ndarray): Output after convolution
            Input_col (numpy.ndarray): retiled and stacked input
        '''
        assert len(Input.shape) == 4 and len(kernel.shape) == 4
        n_x, d_x, h_x, w_x = Input.shape
        n_filter, d_filter, h_filter, w_filter = kernel.shape
        assert d_x == d_filter, "inputs not alligned for standard convolution"

        # check for validity of convolution
        h_out = (h_x - h_filter + 2 * padding) / stride + 1
        w_out = (w_x - w_filter + 2 * padding) / stride + 1

        if not h_out.is_integer() or not w_out.is_integer():
            raise Exception('Invalid output dimension!')

        h_out, w_out = int(h_out), int(w_out)

        Input_col = im2col.im2col_indices(Input, h_filter,
                                          w_filter, padding=padding,
                                          stride=stride)
        kernel_row = kernel.reshape(n_filter, -1)
        out = kernel_row @ Input_col + bias
        out = out.reshape(n_filter, h_out, w_out, n_x)
        out = out.transpose(3, 0, 1, 2)
        return out, Input_col

    def _get_config_(self):
        '''
        returns the dict of params required to recreate the layer
        '''
        outChannels, inChannels, filter_size, _ = self.W.shape
        config = {"outChannels": outChannels, "inChannels": inChannels,
                  "stride": self.stride, "padding": self.padding,
                  "filter_size": filter_size, "trainable": self.trainable}
        base_config = super(Conv2D, self)._get_config_()
        return dict(list(base_config.items())+list(config.items()))


class dilated_Conv2D(Conv2D):
    '''
    2D Dilated Convolutional Layer
    '''

    def __init__(self, outChannels, inChannels, filter_size, dilation=2,
                 stride=1, padding=0, *, init_method='Xavier',
                 trainable=False):
        '''
        Initializes the Convolutional layer
            W is initialized with either Xavier or He initialization
            b is initialized to zero
            dm generates the dilation matrix that is used for dilations

        Args:
            outChannels (int): Number of output channels
            inChannels (int): size of output requred
            filter_size (int): Size of each kernel (filter_size x filter_size)
            dilation (int): Dilation factor to be used
            stride (int): Stride to be used
            padding (int): Padding to be used for convolution
            init_method (str): initialization method to be used for Weights
            trainable (bool): False -> the layer are frozen
                              True -> updated during optimizer step
        '''
        super(dilated_Conv2D, self).__init__(
            outChannels=outChannels, inChannels=inChannels,
            filter_size=filter_size, stride=stride, padding=padding,
            trainable=trainable, init_method=init_method
        )
        assert isinstance(dilation, int) and dilation > 0
        # currently supports only symmetical dilations
        self.dilation = dilation
        self.dm = self._create_dilation_mat_()
        self.l_name = 'dilated_Conv2D'

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the dialted Convolution 2D layer:
        Convolves Inputs with the kernels after dilation

        Args:
            X (numpy.ndarray): Input array
            train (bool): Set true to cache X and X_col

        Returns:
            Output (numpy.ndarray): Output after Convolution
        '''
        # dilate the kernels using dilation matrix
        self.W_exp = self.dm@self.W@self.dm.T
        output, _ = self._convolve_(Input=X, kernel=self.W_exp, bias=self.b,
                                    padding=self.padding, stride=self.stride)
        if train:
            self.cache_in = X
        return output

    def backward(self, dY):
        '''
        Performs a backward pass through the dilated Conv2D Layer

        Args:
            dY (numpy.ndarray): Output gradient backpropagated

        Returns:
            dX (numpy.ndarray): Input grad after backprop
            var_grad_list (list):
                :trainable = True: [(W,dW), (b,db)]
                :trainable = False: [ ]
        '''
        if self.cache_in is None:
            raise RuntimeError(
                f'Gradient cache not defined. When training '
                f'the train argument must be set to true in the forward pass.'
            )
        X = self.cache_in

        n_filter, d_filter, h_filter, w_filter = self.W.shape

        exchange_mat = np.rot90(np.eye(h_filter))
        reversed_w = exchange_mat@self.W@exchange_mat.T
        dialated_reversed_w = self.dm@ reversed_w @self.dm.T

        dX, _ = self._convolve_(Input=dY, kernel=dialated_reversed_w.transpose(
            1, 0, 2, 3), padding=dialated_reversed_w.shape[2]-1)
        assert X.shape == dX.shape

        if self.trainable:
            db = np.sum(dY, axis=(0, 2, 3)).reshape(n_filter, -1)
            dW, _ = self._convolve_(Input=X.transpose(1, 0, 2, 3),
                                    kernel=dY.transpose(1, 0, 2, 3),
                                    stride=self.dilation, padding=0)
            dW = dW.transpose(1, 0, 2, 3)

            assert self.W.shape == dW.shape
            assert self.b.shape == db.shape
            return dX, [(self.W, dW), (self.b, db)]
        # if not trainable
        return dX, []

    def _create_dilation_mat_(self):
        '''
        private:
        generates a dilation matrix that is used to dilate the kernel

        Returns:
            dilation_mat (np.ndarray) : Matrix that is used for dilation
        '''
        I_ = np.eye(self.W.shape[2])
        z = np.zeros((1, self.W.shape[2]))
        res = []
        for i in range(self.W.shape[2]):
            res.append(I_[i])
            for k in range(self.dilation-1):
                res.append(z)
        res = np.row_stack(res)
        dilation_mat = res[:-self.dilation+1]
        return dilation_mat

    def _get_config_(self):
        '''
        returns the dict of params required to recreate the layer
        '''
        outChannels, inChannels, filter_size, _ = self.W.shape
        config = {"dilation": self.dilation}
        base_config = super(Conv2D, self)._get_config_()
        return dict(list(base_config.items())+list(config.items()))

    def __repr__(self):
        return (f'Conv2D Layer with {self.W.shape[0]} '
                f'dilation = {self.dilation} number of filters of '
                f'shape {self.W.Shape[1:]}, Stide = {self.stride}, '
                f'padding = {self.padding} Trainable = {self.trainable}')



# coding=utf-8

from setuptools import find_packages
from setuptools import setup

with open("README.md", 'r') as fh:
    long_description = fh.read()


setup(
    name='simflow',
    version='0.0.1',
    description='simflow',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
    ],
    url='https://github.com/00arun00/SimFlow',
    author="Arun Joseph",
    author_email="arunjoseph.eng@gmail.com"
)



import simflow as sf


def main():
    # load data
    Data, Labels = sf.utils.data_loader_mnist.load_normalized_mnist_data_flat()

    inp_dim = 784
    num_classes = 10
    # create network
    net = sf.model.Model()
    net.add_layer(sf.layers.layers.Dense(inp_dim, 200))
    net.add_layer(sf.layers.activations.ReLU())
    net.add_layer(sf.layers.layers.BN_mean(200))
    net.add_layer(sf.layers.layers.Dense(200, num_classes))

    # add loss function
    net.set_loss_fn(sf.losses.SoftmaxCrossEntropyLoss())

    # add optimizer
    net.set_optimizer(sf.optimizers.SGD(lr=0.01, momentum=0.9, nestrov=True))

    # add iterator
    net.set_iterator(sf.iterators.minibatch_iterator())

    # fit the training data for 5 epochs
    net.fit(Data['train'], Labels['train'], epochs=5)

    # pring scores after training
    print("Final Accuracies after training :")
    print("Train Accuracy: ", net.score(Data['train'],
                                        Labels['train'])[1], end=" ")
    print("validation Accuracy: ", net.score(Data['val'],
                                             Labels['val'])[1], end=' ')
    print("Test Accuracy: ", net.score(Data['test'], Labels['test'])[1])


if __name__ == '__main__':
    main()



import numpy as np
from simflow.layers.convolutional import Conv2D
from simflow.utils.grad_check_utils import numerical_gradient_array
from simflow.utils.im2col import im2col_indices
from unittest import TestCase


class TestConv(TestCase):
    @staticmethod
    def standard_forward(X, w, b, padding, stride):
        n_x, d_x, h_x, w_x = X.shape
        n_filters, d_filter, h_filter, w_filter = w.shape
        h_out = (h_x - h_filter + 2 * padding) // stride + 1
        w_out = (w_x - w_filter + 2 * padding) // stride + 1
        X_col = im2col_indices(X, h_filter, w_filter,
                               padding=padding, stride=stride)
        W_col = w.reshape(n_filters, -1)
        out = W_col @ X_col + b
        out = out.reshape(n_filters, h_out, w_out, n_x)
        out = out.transpose(3, 0, 1, 2)
        return out

    def test_conv_layer_forward_prop(self):
        eps = 1e-8

        batch_size = 32
        filter_size = 3
        h_x, w_x = 7, 7
        inChannels = 3
        n_filter = 5
        padding = 1
        stride = 1

        x = np.random.randn(batch_size, inChannels, h_x, w_x)
        w = np.random.randn(n_filter, inChannels, filter_size, filter_size)
        b = np.random.randn(n_filter, 1)

        c_layer = Conv2D(inChannels=inChannels, outChannels=n_filter,
                         filter_size=filter_size, stride=stride,
                         padding=padding, trainable=True)
        c_layer.W = w
        c_layer.b = b

        out = c_layer.forward(x)
        out_standard = self.standard_forward(x, w, b, padding, stride)

        assert np.allclose(out, out_standard, atol=eps)

    def test_conv_layer_back_prop(self):
        eps = 1e-7

        batch_size = 32
        filter_size = 3
        h_x, w_x = 7, 7
        inChannels = 3
        n_filter = 5
        padding = 1
        stride = 1

        h_out = (h_x - filter_size + 2*padding)//stride + 1
        w_out = (w_x - filter_size + 2*padding)//stride + 1

        x = np.random.randn(batch_size, inChannels, h_x, w_x)
        w = np.random.randn(n_filter, inChannels, filter_size, filter_size)
        b = np.random.randn(n_filter, 1)
        dout = np.random.randn(batch_size, n_filter, h_out, w_out)

        c_layer = Conv2D(inChannels=inChannels, outChannels=n_filter,
                         filter_size=filter_size, stride=stride,
                         padding=padding, trainable=True)
        c_layer.W = w
        c_layer.b = b

        dx_num = numerical_gradient_array(c_layer.forward, x, dout, h=eps)
        dw_num = numerical_gradient_array(lambda w: c_layer.forward(x),
                                          w, dout, h=eps)
        db_num = numerical_gradient_array(lambda b: c_layer.forward(x),
                                          b, dout, h=eps)

        dx, grads = c_layer.backward(dout)
        dw, db = grads[0][1], grads[1][1]

        assert np.allclose(dx, dx_num, atol=eps)
        assert np.allclose(dw, dw_num, atol=eps)
        assert np.allclose(db, db_num, atol=eps)



# source https://github.com/parasdahal/deepnet
import numpy as np


def numerical_gradient_array(f, x, df, h=1e-5):
    """
    Evaluate a numeric gradient for a function that accepts a numpy
    array and returns a numpy array.

    Args:
        f (function): function that is passed to compute gradient for
        x (numpy.ndarray): input to function
        df (numpy.ndarray): output gradient
        h (float): delta around which gradient is calculated

    Returns:
        grad (numpy.ndarray): computed numeric gradient
    """
    grad = np.zeros_like(x)
    it = np.nditer(x, flags=['multi_index'], op_flags=['readwrite'])
    while not it.finished:
        ix = it.multi_index
        oldval = x[ix]
        x[ix] = oldval + h
        pos = f(x).copy()
        x[ix] = oldval - h
        neg = f(x).copy()
        x[ix] = oldval

        grad[ix] = np.sum((pos - neg) * df) / (2 * h)

        it.iternext()
    return grad



from simflow.layers.layer_class import Layer
import numpy as np


class ReLU(Layer):
    '''
    RelU layer
    Represent a nonlinear transformation Y = max(0,X)
    '''

    def __init__(self, *, trainable=True):
        '''
        Initialization :
            Does nothing since nothing to initialize
        '''
        self.cache_in = None
        self.trainable = trainable
        self.l_name = 'ReLU'
        self.params = []

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the ReLU Layer

        Args:
            :X (numpy.ndarray): Input array
            :train (bool): Set to True to enable gardient caching for backward step

        Returns:
            :Out (numpy.ndarray): Output after applying transformation Y = max(0,X)
        '''
        out = np.maximum(X, 0)
        if train:
            self.cache_in = X
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the ReLU Layer

        Args:
            :dY (numpy.ndarray): Output gradient backpropagated from layers in the front

        Returns:
            :dX (numpy.ndarray): Input gradient after backpropagating dY through ReLU layer
            :var_grad_list (list): [], since it has no parameter to be learned
        '''
        if self.cache_in is None:
            raise RuntimeError(
                'Gradient cache not defined. When training the train argument must be set to true in the forward pass.')
        dX = dY*(self.cache_in >= 0)
        return dX, []


# adding aliases
relu = ReLU


class Sigmoid(Layer):
    '''
    Sigmoid layer
    Represent a nonlinear transformation Y = 1/(1+e^(-X))
    '''

    def __init__(self, *, trainable=True):
        '''
        Initialization:
            Does nothing since nothing to initialize
        '''
        self.cache_in = None
        self.trainable = trainable
        self.l_name = 'Sigmoid'
        self.params = []

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the Sigmoid Layer

        Args:
            :X (numpy.ndarray): Input array
            :train (bool): Set to True to enable caching for backward step

        Returns:
            :Out (numpy.ndarray): Output after applying transformation Y = 1/(1+e^(-X))
        '''
        out = 1/(1+np.exp(-X))
        if train:
            self.cache_in = out
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the Sigmoid Layer

        Args:
            :dY (numpy.ndarray): Output gradient backpropagated from layers in the front

        Returns:
            :dX (numpy.ndarray): Input gradient after backpropagating dY through Sigmoid layer
            :var_grad_list (list): [], since it has no parameter to be learned
        '''
        if self.cache_in is None:
            raise RuntimeError(
                'Gradient cache not defined. When training the train argument must be set to true in the forward pass.')
        out = self.cache_in
        dX = dY*(out*(1-out))
        return dX, []


class Tanh(Layer):
    '''
    Tanh layer
    Represent a nonlinear transformation Y = (1-e^(-2X))/(1+e^(-2X)) {tanh}
    '''

    def __init__(self, *, trainable=True):
        '''
        Initialization :
            Does nothing since nothing to initialize
        '''
        self.cache_in = None
        self.trainable = trainable
        self.l_name = 'Tanh'
        self.params = []

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the Tanh Layer

        Args:
            :X (numpy.ndarray): Input array
            :train (bool): Set to True to enable caching for backward step

        Returns:
            :Out (numpy.ndarray): Output after applying transformation Y = tanh(X)
        '''
        out = np.tanh(X)
        if train:
            self.cache_in = out
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the Tanh Layer

        Args:
            :dY (numpy.ndarray): Output gradient backpropagated from layers in the front

        Returns:
            :dX (numpy.ndarray): Input gradient after backpropagating dY through Tanh layer
            :var_grad_list (list): [], since it has no parameter to be learned
        '''

        if self.cache_in is None:
            raise RuntimeError(
                'Gradient cache not defined. When training the train argument must be set to true in the forward pass.')
        out = self.cache_in
        dX = dY*(1-out**2)
        return dX, []


class LeakyReLU(Layer):
    '''
    LeakyReLU layer
    Represent a nonlinear transformation Y = max(alpha*input,input)

    Args:
        :alpha (float): Leakiness parameter
    '''

    def __init__(self, alpha=0.1, *, trainable=True):
        '''
        Initialization :
            initialize the leakiness parameter
        '''
        self.cache_in = None
        self.trainable = trainable
        self.l_name = 'LeakyReLU'
        self.params = []
        self.alpha = alpha

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the LeakyReLU Layer

        Args:
            :X (numpy.ndarray): Input array
            :train (bool): Set to True to enable caching for backward step

        Returns:
            :Out (numpy.ndarray): Output after applying transformation Y = LeakyReLU(X)
        '''
        out = np.maximum(self.alpha*X, X)
        if train:
            self.cache_in = X
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the LeakyReLU Layer

        Args:
            :dY (numpy.ndarray): Output gradient backpropagated from layers in the front

        Returns:
            :dX (numpy.ndarray): Input gradient after backpropagating dY through LeakyReLU layer
            :var_grad_list (list): [], since it has no parameter to be learned
        '''

        if self.cache_in is None:
            raise RuntimeError(
                'Gradient cache not defined. When training the train argument must be set to true in the forward pass.')
        X = self.cache_in
        grad = 1*(X > 0)+self.alpha*(X <= 0)
        dX = dY*grad
        return dX, []


class Softplus(Layer):
    '''
    Represent a nonlinear transformation Y = log(1+exp(X))
    '''

    def __init__(self, *, trainable=True):
        '''
        Initialization :
            Nothing to initialize
        '''
        self.cache_in = None
        self.trainable = trainable
        self.l_name = 'LeakyReLU'
        self.params = []

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the Softplus Layer

        Args:
            :X (numpy.ndarray): Input array
            :train (bool): Set to True to enable caching for backward step

        Returns:
            :Out (numpy.ndarray): Output after applying transformation Y = Softplus(X)
        '''

        out = np.log(1+np.exp(X))
        if train:
            self.cache_in = X
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the Softplus Layer

        Args:
            :dY (numpy.ndarray): Output gradient backpropagated from layers in the front

        Returns:
            :dX (numpy.ndarray): Input gradient after backpropagating dY through Softplus layer
            :var_grad_list (list): [], since it has no parameter to be learned
        '''

        if self.cache_in is None:
            raise RuntimeError(
                'Gradient cache not defined. When training the train argument must be set to true in the forward pass.')
        X = self.cache_in
        grad = 1/(1+np.exp(-X))
        dX = dY*grad
        return dX, []


class exp(Layer):
    '''
    Represent a nonlinear transformation Y = exp(X)
    '''

    def __init__(self, *, trainable=True):
        '''
        Initialization :
            Nothing to initialize
        '''
        self.cache_in = None
        self.trainable = trainable
        self.l_name = 'LeakyReLU'
        self.params = []

    def forward(self, X, train=True):
        '''
        Performs a forward pass through the exp Layer

        Args:
            :X (numpy.ndarray): Input array
            :train (bool): Set to True to enable caching for backward step

        Returns:
            :Out (numpy.ndarray): Output after applying transformation Y = exp(X)
        '''

        out = np.exp(X)
        if train:
            self.cache_in = X
        return out

    def backward(self, dY):
        '''
        Performs a backward pass through the Softplus Layer

        Args:
            :dY (numpy.ndarray): Output gradient backpropagated from layers in the front

        Returns:
            :dX (numpy.ndarray): Input gradient after backpropagating dY through exp layer
            :var_grad_list (list): [], since it has no parameter to be learned
        '''

        if self.cache_in is None:
            raise RuntimeError(
                'Gradient cache not defined. When training the train argument must be set to true in the forward pass.')
        X = self.cache_in
        grad = np.exp(X)
        dX = dY*grad
        return dX, []
